require 'gdbm'
require 'time'
require 'net/pop'
require 'digest/md5'

#===============================================================================
#
#	連番処理付き GDBM クラス
#
class SqGDBM < GDBM

	@@start = '100'
	@@sym = '#last#'

	def initialize(*params)
		super
		self[@@sym] = @@start unless(self[@@sym])
	end

	#-----------------------------------------------------------
	#
	#	最後の連番を返す
	#
	def last_sq
		self[@@sym]
	end

	#-----------------------------------------------------------
	#
	#	加算した連番を返す
	#
	def inc_sq
		self[@@sym] = (self[@@sym].to_i + 1).to_s
	end
end

#===============================================================================
#
#	モデルが自分の関連ビューを管理するクラス
#
class MaveModelViews < Hash

	def initialize
		@sq = 0
	end

	def []=(name, view)
		super(name || (@sq += 1).to_s, view)
	end

	def update													# モデル側の更新を契機として自発的に
		each {|key, value|
			value.update										# キー操作とは非同期にビューを再描画
		}
	end
end

#===============================================================================
#
#	モデルベースクラス
#
class MaveBaseModel

	def initialize(params)
		@configs = params[:CONFIGS]
		@views = MaveModelViews.new
		@clean = 0; @dirty = 1									# 情報更新の有無の管理
	end

	def tie(view, name = nil)
		@views[name] = view
	end

	def dirty?
		@clean < @dirty and @clean = @dirty
	end
end

#===============================================================================
#
#	メールアカウントモデルクラス
#
class MaveAccount < MaveBaseModel

	attr_reader :name
	attr_reader :enable
	attr_reader :pop_server
	attr_reader :hash_id

	def initialize(params)
		super
		@account		= params[:ACCOUNT]

		@name			= @account[:NAME]
		@enable			= @account[:ENABLE]

		@pop_server		= @account[:POP_SERVER]
		@pop_port		= @account[:POP_PORT]
		@pop_account	= @account[:POP_ACCOUNT]
		@pop_password	= @account[:POP_PASSWORD]
		@pop_keep_time	= @account[:POP_KEEP_TIME]

		@hash_id		= Digest::MD5.hexdigest(@account[:USER_ADDRESS])[0, 8]
		@pop_uids		=   GDBM.new(@configs[:ROOT_DIRECTORY] + "/pop_uids_#{name}")
	end

	#-----------------------------------------------------------
	#
	#	アカウントの各要素を返す
	#
	def [](param)
		@account[param]
	end

	#-----------------------------------------------------------
	#
	#	POP する
	#
	def pop
		now = Time.new.to_i
		Net::POP3.start(@pop_server, @pop_port, @pop_account, @pop_password) {|pop|
			yield('Connected.')
			popmails = []
			pop.mails.each {|popmail|
				if(@pop_uids[popmail.unique_id])				# 既知のメール
					if(@pop_uids[popmail.unique_id].to_i < now - @pop_keep_time)
						popmail.delete
						@pop_uids.delete(popmail.unique_id)
					end
				else											# 未知のメール
					popmails << popmail
				end
			}
			yield(popmails.size)
			popmails.each {|popmail|
				yield(popmail)
				@pop_uids[popmail.unique_id] = now.to_s
				@dirty += 1
			}
			@pop_uids.keys.each {|unique_id|
				@pop_uids.delete(unique_id) if(@pop_uids[unique_id].to_i < now - @pop_keep_time)
			}
		}
	end
end

#===============================================================================
#
#	メールアカウント「群」モデルクラス
#
class MaveAccounts < MaveBaseModel

	def initialize(params)
		super
		@accounts		= {}

		@regular_nth = nil
		@configs[:ACCOUNTS].each_index {|nth|
			it = @accounts[@configs[:ACCOUNTS][nth][:NAME]] = MaveAccount.new({:CONFIGS => @configs, :ACCOUNT => @configs[:ACCOUNTS][nth]})
			@regular_nth = nth if(!@regular_nth and it.enable)
		}
		unless(@regular_nth)
			print "#{_('Prepare one or more enable accounts [mave.config].')}\n"
			exit
		end
	end

	#-----------------------------------------------------------
	#
	#	メールアカウントを順に渡す(有効なアカウントのみ)
	#
	def each													#### 適切な順序で返すようにする
		@accounts.each {|name, account|
			yield(account) if(account.enable)
		}
	end

	#-----------------------------------------------------------
	#
	#	既定のメールアカウントを返す／変更する
	#
	def regular
		@accounts[@configs[:ACCOUNTS][@regular_nth][:NAME]]
	end
	def previous
		begin
			@regular_nth -= 1; @regular_nth %= @configs[:ACCOUNTS].size
		end until(@accounts[@configs[:ACCOUNTS][@regular_nth][:NAME]].enable)
		regular
	end
	def next
		begin
			@regular_nth += 1; @regular_nth %= @configs[:ACCOUNTS].size
		end until(@accounts[@configs[:ACCOUNTS][@regular_nth][:NAME]].enable)
		regular
	end
end

#===============================================================================
#
#	ディレクトリモデルクラス
#
class MaveDirectory < MaveBaseModel

	attr_reader :path											# ディレクトリの絶対パス

	def initialize(params)
		super
		@path			= params[:PATH]

		@cluster_dirs	= @configs[:CLUSTER_DIRS]
		@cluster_ext	= @configs[:CLUSTER_EXT]

		unless(File.directory?(@path))							# (クラスタ)ディレクトリ作成
			Dir.mkdir(@path)
			(0...@cluster_dirs).each {|nth|
				Dir.mkdir('%s/%04d.%s' % [@path, nth, @cluster_ext])
			}
		end

		@cluster_paths	= []									# クラスタディレクトリ認識
		Dir.glob(@path + '/*.' + @cluster_ext) {|dir|
			@cluster_paths << File.basename(dir) + '/'
		}
		@cluster_paths << '' if(@cluster_paths.empty?)

		@filename_sq	= SqGDBM.new(@path + '/filename_sq')	# ベースファイル名<-メール連番
	end

	#-----------------------------------------------------------
	#
	#	メールファイルをディレクトリ内に新規作成
	#
	def create_mailfile
		halfname = '%s%08d.eml' % [@cluster_paths[(it = @filename_sq.inc_sq.to_i) % @cluster_paths.size], it]
		File.open(@path + '/' + halfname, 'w') {|fh|
			yield(fh)
		}
		@dirty += 1
		halfname
	end

	#-----------------------------------------------------------
	#
	#	メールファイルを上書き
	#
	def overwrite_mailfile(halfname)
		File.open(@path + '/' + halfname, 'w') {|fh|
			yield(fh)
		}
		@dirty += 1
		halfname
	end

	#-----------------------------------------------------------
	#
	#	メールファイルのハッシュ値を返す
	#
	def md5(halfname)
		Digest::MD5.file(@path + '/' + halfname).to_s
	end

	#-----------------------------------------------------------
	#
	#	メールファイルを削除
	#
	def delete(halfname)
		File.delete(@path + '/' + halfname)
		@dirty += 1
	end

	#-----------------------------------------------------------
	#
	#	最後の連番を返す
	#
	def last_sq
		@filename_sq.last_sq
	end
end

#===============================================================================
#
#	メールフォルダモデルクラス
#
class MaveFolder < MaveDirectory

	attr_reader :name											# フォルダの名前

	def initialize(params)
		params[:PATH]	= params[:CONFIGS][:ROOT_DIRECTORY] + '/' + params[:NAME]
		super
		@name			= params[:NAME]

		@sq_rootsq		= SqGDBM.new(@path + '/sq_rootsq')		# メール連番<-最新順ルート連番

		@messageid_sq	=   GDBM.new(@path + '/messageid_sq')	# メッセージID<-メール連番
		@sq_messageid	=   GDBM.new(@path + '/sq_messageid')	# メール連番<-メッセージID

		@rootsq_sq		=   GDBM.new(@path + '/rootsq_sq')		# ルートメール連番<-メール連番
		@parentsq_sq	=   GDBM.new(@path + '/parentsq_sq')	# 親メール連番<-メール連番
		@childsq_sq		=   GDBM.new(@path + '/childsq_sq')		# 子メール連番群<-メール連番

		# フォルダの設定ファイルを読む(config と bind? メソッドが追加される)
		begin
			File.open(@path + '/mave.folderconfig') {|fh|
				instance_eval(fh.read)
			}
		rescue													# 設定ファイルが無ければ適当
			def folder_config
				{}
			end
			def bind?(mail)
				name != 'Inbox' ? false : true					####
			end
		end
	end

	#-----------------------------------------------------------
	#
	#	指定のメール連番のメールインスタンスを返す
	#
	def get_mail(sq)
		sq ? MaveMail.new({:CONFIGS => @configs, :FILE => File.new(@path + '/' + @filename_sq[sq]), :FOLDER => self, :SQ => sq}) : nil
	end
 
	#-----------------------------------------------------------
	#
	#	メールファイルのハッシュ値を返す
	#
	def md5(sq)
		super(@filename_sq[sq])
	end

	#-----------------------------------------------------------
	#
	#	フォルダ内にメールを追加する
	#
	def add_mail(mail)
		filename = create_mailfile {|fh|
			mail.rewind
			mail.each {|line|
				fh.write(line)									# ファイルコピー
			}
		}
		@filename_sq[sq = @filename_sq.last_sq] = filename
		@messageid_sq[sq] = mail.message_id
		@sq_messageid[mail.message_id] = sq

		# 直近の親を捜す
		parent_sqi = 0; it = nil
		mail.each_in_reply_to {|id|
			parent_sqi = it.to_i if(it = @sq_messageid[id] and it.to_i > parent_sqi)
		}
		mail.each_reference {|id|
			parent_sqi = it.to_i if(it = @sq_messageid[id] and it.to_i > parent_sqi)
		} unless(parent_sqi)

		# 親子関係を登録する
		if(parent_sqi == 0)										# 親が不在->自分がルート
			@sq_rootsq[rootsq = @sq_rootsq.inc_sq] = @rootsq_sq[sq] = sq
			@parentsq_sq[sq] = rootsq							# ルートの親はrootsqを指す
		else
			parent_sq = @parentsq_sq[sq] = parent_sqi.to_s		# 自分の親、自分の親の子を登録
			@childsq_sq[parent_sq] = '' unless(@childsq_sq[parent_sq])
			@childsq_sq[parent_sq] = @childsq_sq[parent_sq].split(',').push(sq).join(',')
			root_sq = @rootsq_sq[sq] = @rootsq_sq[parent_sq]	# 自分のルートは、親のルートと同じ
			unless((old_rootsq = @parentsq_sq[root_sq]) == @sq_rootsq.last_sq)
				@sq_rootsq[rootsq = @sq_rootsq.inc_sq] = root_sq	# 自分のルートを最新化
				@parentsq_sq[root_sq] = rootsq
				@sq_rootsq.delete(old_rootsq)
			end
		end
		@dirty += 1
	end

	#-----------------------------------------------------------
	#
	#	フォルダ内にメールを上書きする
	#
	def overwrite_mail(mail, old_mail)
		filename = overwrite_mailfile(@filename_sq[old_mail.sq]) {|fh|
			mail.rewind
			mail.each {|line|
				fh.write(line)									# ファイルコピー
			}
		}
#		@messageid_sq[sq] = mail.message_id						#### メッセージ ID 等、変化するかも
#		@sq_messageid[mail.message_id] = sq
		@dirty += 1
	end

	#-----------------------------------------------------------
	#
	#	フォルダ内のメールを返す
	#
	def next_sq(targetsq)										# 次を返す
		each_sq(targetsq) {|sq, depth|
			break sq unless(sq == targetsq)
		}
	end

	def each_sq(startsq = nil)									# 次を順に渡す
		unless(startsq)
			rootsqi = @sq_rootsq.last_sq.to_i
			started = true
		else
			rootsqi = @parentsq_sq[@rootsq_sq[startsq]].to_i
			started = false
		end
		while(rootsqi > 0)
			if(sq = @sq_rootsq[rootsqi.to_s])
				each_sq2(sq, 0) {|sq2, depth2|
					started = true if(!started and sq2 == startsq)
					yield(sq2, depth2) if(started)
				}
			end
			rootsqi -= 1
		end
		nil
	end

	def each_sq2(sq, depth)										# 再帰的に子を順に渡す
		yield(sq, depth)
		@childsq_sq[sq].split(',').each {|childsq|
			each_sq2(childsq, depth + 1) {|sq2, depth2|
				yield(sq2, depth2)
			}
		} if(@childsq_sq[sq])
	end

	def reverse_each_sq(startsq = nil)							# 前を順に渡す
		unless(startsq)
			rootsqi = 0
			started = true
		else
			rootsqi = @parentsq_sq[@rootsq_sq[startsq]].to_i
			started = false
		end
		max_rootsqi = @sq_rootsq.last_sq.to_i; buf = []; while(rootsqi <= max_rootsqi)
			if(sq = @sq_rootsq[rootsqi.to_s])
				buf.clear; each_sq2(sq, 0) {|sq2, depth2|
					buf << [sq2, depth2]
				}
				buf.reverse.each {|sq2, depth2|
					started = true if(!started and sq2 == startsq)
					yield(sq2, depth2) if(started)
				}
			end
			rootsqi += 1
		end
	end

	def previous_sq(targetsq)									# 前を返す
		reverse_each_sq(targetsq) {|sq, depth|
			break sq unless(sq == targetsq)
		}
	end
end

#===============================================================================
#
#	メールフォルダ「群」モデルクラス
#
class MaveFolders < MaveBaseModel

	def initialize(params)
		super
		@folders		= {}

		begin
			Dir.open(@configs[:ROOT_DIRECTORY]) {|dir|
				dir.each {|name|
					next if(name =~ /^[._]/ or !File.directory?(dir.path + '/' + name))
					@folders[name] = MaveFolder.new({:CONFIGS => @configs, :NAME => name})
				}
			}
		rescue
			print "#{_('Prepare the root directory [%s].') % @configs[:ROOT_DIRECTORY]}\n"
			exit
		end
	end

	#-----------------------------------------------------------
	#
	#	メールフォルダ「群」を通して、メールフォルダを作成／開く
	#
	def open_folder(name)
		@folders[name] || (@folders[name] = MaveFolder.new({:CONFIGS => @configs, :NAME => name}))
	end

	#-----------------------------------------------------------
	#
	#	メールフォルダを順に渡す
	#
	def each													#### 適切な順序で返すようにする
		@folders.each {|name, folder|
			yield(folder)
		}
	end
end

#===============================================================================
#
#	メールクラス
#
class MaveMail < MaveBaseModel

	attr_reader :folder
	attr_reader :sq

	@@debug_what_charset = false

	#-----------------------------------------------------------
	#
	#	キャラクタセット情報表示、切り替え(デバッグ用)
	#
	def self.toggle_what_charset
		@@debug_what_charset = !@@debug_what_charset
	end

	def initialize(params)
		super
		@file			= params[:FILE]
		@boundary		= params[:BOUNDARY]
		@folder			= params[:FOLDER]
		@sq				= params[:SQ]

		parse_header
		parse_body
	end

	def path
		@file.path
	end

	def pos
		@file.pos
	end

	def rewind
		@file.rewind											#### とりあえず
	end

	def each
		@file.each {|line| yield line }							#### とりあえず
	end

	def md5
		Digest::MD5.file(@file.path).to_s
	end

	#-----------------------------------------------------------
	#
	#	メールヘッダを解析
	#
	def parse_header
		@header = {}; last = ''
		@file.each {|line|
			line.chomp!
			if(line =~ /^(\S+?):\s*(.*)/)
				@header[last = $1.capitalize] = $2
			elsif(line =~ /^\s+(=\?.*)/)						# encode 部分が連続した場合の特別扱い
				@header[last] << $1
			elsif(line =~ /^\s+(.*)/)
				@header[last] << " #{$1}"
			elsif(line =~ /^\s*$/)
				break
			else
				raise "Mail format error '#{line.inspect}'"
			end
		}
	end

	def parse_body
		@content = {
			'type'				=> {
				'type'				=> 'text/plain',
				'param'				=> {
					'charset'			=> 'us-ascii',
				},
			},
			'transfer-encoding'	=> {
				'type'				=> '7bit',
			},
			'disposition'		=> {
				'type'				=> nil,
				'param'				=> {
				},
			},
		}
		#
		#	Content-type, Content-transfer-encoding
		#		http://tools.ietf.org/html/rfc2045
		#
		#	Content-Disposition
		#		http://tools.ietf.org/html/rfc1806
		#
		['type', 'transfer-encoding', 'disposition'].each {|shalf|
			if(@header[it = 'Content-' + shalf])
				qss = []
				tps = @header[it].gsub(/"([^"]*)"/) { qss << $1; '"' }.split(';')
				@content[shalf]['type'] = tps.shift.downcase
				tps.each {|ps|
					param = @content[shalf]['param']
					param[$1.downcase] = $2 if(ps =~ %r|(\S+)=(\S+)|)
					param[$1.downcase] = qss.shift if($2 == '"')
				}
			end
		}

#### Mixed
#### Alternative
#### Parallel
#### Digest	-> 中のデフォルト content-type を替える必要あり？

		it = MaveMailParts[@content['type']['type']] || MaveMailParts['unknown']
		@body = it.new(@file, @content, @boundary)
	end

	#-----------------------------------------------------------
	#
	#	メールヘッダの行を順に渡す
	#
	def headers_each
		@file.rewind
		@file.each {|line|
			break if(line =~ /^\s*$/)
			yield(line.chomp)
		}
	end

	#-----------------------------------------------------------
	#
	#	メールボディの行を順に渡す
	#
	def body_each
		nth = -1; while(it = self[nth += 1])
			yield(it)
		end
	end

	#-----------------------------------------------------------
	#
	#	メールの特定の行を返す
	#
	def [](nth)
		if(@body[nth] and @body[nth].is_a?(String))
			(@@debug_what_charset ? @content['type']['param']['charset'] + ': ' : '') + \
			@body[nth].decode_cs('UTF-8', @content['type']['param']['charset'])
		else
			@body[nth]
		end
	end

	#-----------------------------------------------------------
	#
	#	メールの件名、送信者、宛先、同報、送信時刻などを返す
	#
	def subject
		@header['Subject'] || '--no subject--'
	end
	def from
		@header['From'] || '--no from--'
	end
	def to
		@header['To'] || nil
	end
	def cc
		@header['Cc'] || nil
	end
	def date
		Time.rfc2822(@header['Date']) rescue nil
	end

	#-----------------------------------------------------------
	#
	#	Message-ID を返す
	#
	def message_id
		@message_id || (@header['Message-id'] =~ /<[^>]+?>/ ? (@message_id = $&) : raise('Bad Message-ID.'))
	end

	#-----------------------------------------------------------
	#
	#	In-Reply-To, References を返す
	#
	def in_reply_to
		@header['In-reply-to'] || nil
	end
	def references
		@header['References' ] || nil
	end
	def each_in_reply_to
		(it = @header['In-reply-to']) and it.scan(/<[^>]+?>/) {|id| yield id }
	end
	def each_reference
		(it = @header['References']) and it.scan(/<[^>]+?>/) {|id| yield id }
	end

	#-----------------------------------------------------------
	#
	#	メールのサイズを返す
	#
	def size
		@file.lstat.size
	end

	#-----------------------------------------------------------
	#
	#	マルチパートかどうかを返す
	#
	def multipart?
		@content['type']['type'].index('multipart')
	end

	#-----------------------------------------------------------
	#
	#	メール情報を返す
	#
	def identify
		yield([_('Message-ID: %s'), message_id])
		yield([_('  FilePath: %s'), path])
		yield([_('  FileSize: %d'), size])
	end

	#-----------------------------------------------------------
	#
	#	メールの概要を返す
	#
	def abstract(sq, padding = '')
		'%s%s %3d %-*s %*s %4s %c%s %s%s' % [
			'.',	#### red?  ? '.' : 'x',
			'.',	#### notice? ? '#' : (flag? ? 'F' : '.'),
			(it = @folder.last_sq.to_i - sq.to_i) < 1000 ? it : 999,
			fw = 16,
			from.decode_mh.snip(fw),
			dw = Time.mystrftime_len,
			date ? date.mystrftime : '-' * dw,
			size.to_h,
			?.,		#### '.vw'[toyou?],
			multipart? ? '@' : '.',
			padding,
			(padding.size == 0 ? '+' : '-') + subject.decode_mh,
		]
	end
end

#===============================================================================
#
#	ダミーメールもどきクラス
#
class MavePseudoMailFile

	def each
		yield("\n")
	end

	def pos
		0
	end
end

#===============================================================================
#
#	メールもどきクラス
#
class MavePseudoMail < MaveMail

	def initialize(params)
		params[:FILE]	= MavePseudoMailFile.new unless(params[:FILE])
		super
		@account		= params[:ACCOUNT]
		@mail			= params[:MAIL]
		@formtype = {
			:NEW			=> method(:new_message_each),
			:RENEW			=> method(:renew_message_each),
			:REPLY			=> method(:reply_message_each),
			:REPLY_TO_ALL	=> method(:reply_to_all_message_each),
			:FORWARD		=> method(:forward_message_each),
			:RESEND			=> method(:resend_message_each),
			:EDIT			=> method(:edit_message_each),
		}
		@each_func		= @formtype[params[:MODE]] || nil
	end

	def parse_header
		super
		@body_pos = pos
	end

	#-----------------------------------------------------------
	#
	#	メールもどきを正しいメール形式で順に渡す
	#
	def each
		super {|line|
			if(pos <= @body_pos)
				if(line =~ /^Date: /)
					yield("Date: #{Time.now.rfc2822}\n")		# Date は現時刻に変更
				else
					yield(line.chomp + "\n")					# ヘッダを仲介
				end
			else
				yield((line.chomp + "\n").tojis)				# ボディを仲介
			end
		}
	end

	#-----------------------------------------------------------
	#
	#	各種メールもどきを順に渡す
	#
	def pseudo_each
		@each_func.call {|line|
			yield(line)
		}
	end

	def headers_each
		yield("From: #{@account[:FROM]}")
		yield("To: #{@account[:DEFAULT_TO] || ''}")
		yield("Cc: #{@account[:DEFAULT_CC] || ''}")
		yield("Bcc: #{@account[:DEFAULT_BCC] || ''}")
		yield("Subject: --#{_('no subject')}--")
		yield("Date: #{Time.now.rfc2822}")
		yield("X-Mailer: #{@configs[:MAILER]}")
		yield("Message-ID: <#{Time.now.to_f.to_s + ".#{@account.hash_id}@mave." + @configs[:HOSTNAME]}>")
		yield('In-Reply-To: ')
		yield('References: ')
		yield('MIME-Version: 1.0')
		yield('Content-Type: text/plain; charset=ISO-2022-JP')	#### アカウントの指定による
		yield('Content-Transfer-Encoding: 7bit')
		yield('X-Mave-Attachments: ')							# 添付用の擬似ヘッダ
	end

	def quote_each
		yield('')
		yield("---- #{_('Original Message')} ----")
		yield("#{_('      From: ')}#{@mail.from.decode_mh}")	if(@mail.from)
		yield("#{_('        To: ')}#{@mail.to.decode_mh}")		if(@mail.to)	#### 長すぎ注意
		yield("#{_('        Cc: ')}#{@mail.cc.decode_mh}")		if(@mail.cc)
		yield("#{_('   Subject: ')}#{@mail.subject.decode_mh}")	if(@mail.subject)
		yield("#{_('      Sent: ')}#{@mail.date.rfc2822}")		if(@mail.date)	#### 好みの形式
		yield("#{_('Message-ID: ')}#{@mail.message_id}")		if(@mail.message_id)
		yield('')
		@mail.body_each {|line|
			line.chomp!
			yield((line =~ /^>/ ? ">#{line}" : "> #{line}").gsub(/[\x00-\x1F]/, '^x'))
		}
	end

	#-----------------------------------------------------------
	#
	#	新規メールもどき作成
	#
	def new_message_each
		headers_each {|line|
			next if(line =~ /^In-Reply-To: /)
			next if(line =~ /^References: /)
			yield(line)
		}
		yield('')
		@account[:GREETING].chomp.split(/\r?\n/, -1).each {|line|
			yield(line.gsub(/[\x00-\x1F]/, ''))
		}
		@account[:SIGNATURE].chomp.split(/\r?\n/, -1).each {|line|
			yield(line.gsub(/[\x00-\x1F]/, ''))
		}
	end

	def renew_message_each
	end

	#-----------------------------------------------------------
	#
	#	返信メールもどき作成
	#
	def reply_message_each
		headers_each {|line|
			if(line =~ /^Subject: /)
				@mail and it = @mail.subject and yield("Subject: #{('Re: ' + it).group_re}")
			elsif(line =~ /^In-Reply-To: /)
				@mail and it = @mail.message_id and yield("In-Reply-To: #{it}")
			elsif(line =~ /^References: /)
				if(@mail)
					refs = []
					@mail.each_reference   {|id| refs |= [id] }
					@mail.each_in_reply_to {|id| refs |= [id] }
					max_refs = 4; refs.slice!(max_refs >> 1, (it = refs.size - max_refs) > 0 ? it : 0)
					it = @mail.message_id and refs |= [it]
					yield("References: #{refs.join(' ')}") if(refs.size > 0)
				end
			else
				yield(line)
			end
		}
		yield('')
		@account[:GREETING].chomp.split(/\r?\n/, -1).each {|line|
			yield(line.gsub(/[\x00-\x1F]/, ''))
		}
		@account[:SIGNATURE].chomp.split(/\r?\n/, -1).each {|line|
			yield(line.gsub(/[\x00-\x1F]/, ''))
		}
		quote_each {|line|
			yield line
		} if(@mail)
	end

	def reply_to_all_message_each
	end

	def forward_message_each
	end

	def resend_message_each
	end

	#-----------------------------------------------------------
	#
	#	メールもどきの(再)編集
	#
	def edit_message_each
		@mail.headers_each {|line|
			yield(line)
		} if(@mail)
		yield('')
		@mail.body_each {|line|
			yield(line.chomp.gsub(/[\x00-\x1F]/, '^x'))
		} if(@mail)
	end
end

#===============================================================================
#
#	プレビューモデルクラス
#
class MavePreview < MaveBaseModel

	def initialize(params)
		super
	end
end

#===============================================================================
#
#	ステータスモデルクラス
#
class MaveStatus < MaveBaseModel

	def initialize(params)
		super
		@logs = []
		@max_logs = 1000
	end

	#-----------------------------------------------------------
	#
	#	ログを追加
	#
	def log(log)
		@logs << sprintf(*log)
		@logs.shift while(@logs.size > @max_logs)
		@dirty += 1
		@views.update											# 表示のリアルタイム更新
	end

	#-----------------------------------------------------------
	#
	#	ログを順に返す
	#
	def recent_each(length, back = 0)
		to = @logs.size - back
		(to - length...to).each {|nth|
			yield(nth < 0 ? nil : @logs[nth])
		}
	end
end

#===============================================================================
#
#	メールパート、ベースクラス
#
class MaveBasePart

	attr_reader :index

	def initialize(file, content, boundary)
		@file = file
		@pos = @file.pos
		@content = content
		@boundary = boundary
	end
end

#===============================================================================
#
#	メールパート、マルチパートクラス
#
class MaveMultipartMixedPart < MaveBasePart

	def initialize(file, content, boundary)
		super
		@mboundary = @content['type']['param']['boundary']
	end

	def [](nth)
		@index = [@pos] and @offset = nth unless(@index)
		wth = oth = nth - @offset
		(it = @index.size - 1) < oth and oth = it
		begin
			if(@index[oth].is_a?(Integer))						# キャッシュはファイルポインタ？
				@file.pos = @index[oth]
				line = @file.gets
				unless(line and line.index("--#{@mboundary}"))
					@index[oth + 1] = @file.pos unless(@index[oth + 1])
				else
					unless(line.index("--#{@mboundary}--"))
						@index[oth] = MaveMail.new({:CONFIGS => @configs, :FILE => @file, :BOUNDARY => @mboundary})
						line = @index[oth][@offset + oth]
						@index[oth + 1] = @index[oth]
					else
						@index[oth] = @file.pos
						line = self[@offset + oth]
					end
				end
			elsif(@index[oth].is_a?(MaveMail))					# キャッシュは MaveMail インスタンス？
				line = @index[oth][@offset + oth]
				unless(line.is_a?(Integer))
					@index[oth + 1] = @index[oth] unless(@index[oth + 1])
				else
					@index[oth] = line
					line = self[@offset + oth]
				end
			end
			oth += 1
		end until(@index[wth])
		line
	end
end

#===============================================================================
#
#	メールパート、プレーンテキストクラス
#
class MaveTextPlainPart < MaveBasePart

	def [](nth)
		encoding = @content['transfer-encoding']['type'].upcase
		if(encoding == 'BASE64' or encoding == 'QUOTED-PRINTABLE')	####
			unless(@index)
				@offset = nth
				lline = ''
				while(line = @file.gets)							# コンテンツ部分を丸ごと 1 行に結合
					break if(@boundary and line.index(@boundary))
					lline << line									#### チカラ技すぎる
					bpos = @file.pos
				end
				@index = lline.decode_ec(encoding).split(/\r?\n/)
				@index << bpos										# boundary の直前位置を保持
			end
			wth = nth - @offset
			if(@index[wth])
				if(@index[wth].is_a?(String))
					line = @index[wth]
				elsif(@index[wth].is_a?(Integer))					# キャッシュはファイルポインタ？
					line = @boundary ? @index[wth] : nil
				end
			else
				line = nil
			end
			line
		else
			@index = [@pos] and @offset = nth unless(@index)
			wth = oth = nth - @offset
			(it = @index.size - 1) < oth and oth = it
			begin
				@file.pos = @index[oth]
				line = @file.gets
				@index[oth += 1] = @file.pos
			end until(@index[wth])
			(@boundary and line.index(@boundary)) ? @index[wth] : line
		end
	end
end

#===============================================================================
#
#	メールパート、不明コンテンツクラス
#
class MaveUnknownPart < MaveBasePart

	def [](nth)
		unless(@index)
			@index = [] and @offset = nth
			@index << "==\n"
			@index << "== Unknown Content-type [#{@content['type']['type']}] ==\n"
			dtype = @content['disposition']['type'] || 'unknown'
			dfilename = (it = @content['disposition']['param']['filename']) ? it.decode_mh : 'unknown'
			@index << "== Content-Desposition [#{dtype}/#{dfilename}] ==\n"
			@index << "==\n"
			while(line = @file.gets)							# 不明コンテンツ部分を読み飛ばす
				break if(@boundary and line.index(@boundary))
				bpos = @file.pos
			end
			@index << bpos										# boundary の直前に位置
		end
		wth = nth - @offset
		if(@index[wth])
			if(@index[wth].is_a?(String))
				line = @index[wth]
			elsif(@index[wth].is_a?(Integer))					# キャッシュはファイルポインタ？
				line = @boundary ? @index[wth] : nil
			end
		else
			line = nil
		end
		line
	end
end

#===============================================================================
#
#	メールパートの登録
#
MaveMailParts = {
	'multipart/mixed'	=> MaveMultipartMixedPart,
	'text/plain'		=> MaveTextPlainPart,
	'unknown'			=> MaveUnknownPart,
}

__END__

